#pragma once

#include "File.h"
#include "Directory.h"
#include "StringUtils.h"

#include <vector>
#include <string>
#include <iostream>

class TemplateNode
{
public:

	TemplateNode()
	{

	}

	virtual void addChild(std::unique_ptr<TemplateNode> child)
	{
		mChildren.push_back(std::move(child));
	}

protected:
	std::vector<std::unique_ptr<TemplateNode> > mChildren;
};

using TemplateNodePtr = std::unique_ptr<TemplateNode>;

class TemplateBlock : public TemplateNode
{
public:
	TemplateBlock(const std::string& name)
		: mName(name)
	{

	}

	void addLine(const std::string& line)
	{
		mBody.push_back(line);
	}

private:
	std::string mName;
	std::vector<std::string> mBody;
};

class TemplateExpression : public TemplateNode
{
public: 
	TemplateExpression(const std::string& content)
		: mContent(content)
	{

	}
private:
	std::string mContent;
};

class TemplateTextBody : public TemplateNode
{
public:
	TemplateTextBody(const std::string& content)
		: mContent(content)
	{

	}
private:
	std::string mContent;
};

class TemplateFile
{
public:
	TemplateFile(const Path& path)
		: mPath(path)
	{

	}

	std::vector<std::string> getProcessedContent() const
	{
		return mProcessedContent;
	}

	std::string getName() const
	{
		return mPath.stem().string();
	}

	void loadContent()
	{
		std::cout << "Trying to load file at: " << mPath << std::endl;
		mRawContent = File(mPath).readLines();
		mWorkingLine = 0;
		for (const auto& line : mRawContent)
		{
			processLine(line);
			mWorkingLine++;
		}
	}

	void processLine(const std::string& line)
	{
		bool last_was_ldelimiter{ false };
		bool last_was_statement_rdelimiter{ false };
		bool last_was_expression_rdelimiter{ false };
		bool in_statement{ false };
		bool in_expression{ false };
		std::string working_string;
		for (auto c : line)
		{
			if (c == '{')
			{
				if (last_was_ldelimiter)
				{
					in_expression = true;
					last_was_ldelimiter = false;
					working_string = "";
				}
				else
				{
					last_was_ldelimiter = true;
				}			
			}
			else if (c == '%' && last_was_ldelimiter)
			{
				last_was_ldelimiter = false;
				in_statement = true;
				working_string = "";
			}
			else if (c == '%' && in_statement)
			{
				last_was_statement_rdelimiter = true;
			}
			else if (c == '}' && (last_was_statement_rdelimiter || in_expression || last_was_expression_rdelimiter))
			{
				if (last_was_statement_rdelimiter)
				{
					onFoundStatement(working_string);
					last_was_statement_rdelimiter = false;
					working_string = "";
				}
				else if (in_expression)
				{
					last_was_expression_rdelimiter = true;
					in_expression = false;
				}
				else
				{
					onFoundExpression(working_string);
					last_was_expression_rdelimiter = false;
					working_string = "";
				}
			}
			else
			{
				working_string += c;
			}
		}
	}

	void onFoundStatement(const std::string& statement_string)
	{
		std::cout << "Got Statement: " << statement_string << std::endl;
		const auto statement_elements = StringUtils::split(statement_string);

		if (statement_elements.size() == 0)
		{
			return;
		}
		const auto statement_type = statement_elements[0];
		if (statement_type == "block")
		{
			onFoundBlock(statement_elements);
		}
		else if (statement_type == "endblock")
		{
			onFoundEndBlock(statement_elements);
		}

	}

	void onFoundBlock(const std::vector<std::string> args)
	{

	}

	void onFoundEndBlock(const std::vector<std::string> args)
	{

	}

	void onFoundExtends(const std::vector<std::string> args)
	{

	}

	void onFoundExpression(const std::string& expression_string)
	{
		std::cout << "Got Expression: " << expression_string << std::endl;
	}

private:
	Path mPath;
	unsigned mWorkingLine{ 0 };
	std::string mParentName;
	std::vector<std::string> mRawContent;
	std::vector<std::string> mProcessedContent;
	std::vector<TemplateBlock> mBlocks; 
};

class TemplatingEngine
{
public:
	TemplatingEngine(const Path& workingDirectory)
		: mWorkingDirectory(workingDirectory)
	{

	}

	void loadTemplateFiles()
	{
		const auto files = Directory::getFilesWithExtension(mWorkingDirectory, mTemplateExtension);
		for (const auto& file : files)
		{
			mTemplateFiles.push_back(TemplateFile(file));
		}
	}

	void processTemplate(const std::string& name)
	{
		for (auto& file : mTemplateFiles)
		{
			const auto filename = file.getName();
			if (filename == name)
			{
				processTemplate(file);
				break;
			}
		}
	}

	void processTemplate(TemplateFile& file)
	{
		file.loadContent();
	}

	


private:
	std::vector<TemplateFile> mTemplateFiles;
	Path mWorkingDirectory;
	std::string mTemplateExtension{ ".html" };
};